﻿using System;
using System.Collections.Generic;
using System.Linq;
using Unity;
using UnityEngine;

/* This is the unmodified AA player, just in a namespace to avoid conflict.*/

namespace RT_LipSync_Player_VH34
{
    class RT_LipSync : MVRScript
    {
        private AudioSource voice_source;
        AudioSourceControl headAudioSource;

        private DAZMorph[] viseme_morphs;
        private DAZMorph mouthOpen_morph;

        private const float fs = 48e3f;
        private const int no_visemes = 16;
        private const int no_phonemes = 16;
        private const int N_freq_bins = 4096;

        private const float frequency_resolution = fs / N_freq_bins / 2;

        /* General Parameters */
        float bs_change_speed = 0.6f;
        float volume_scaling = 15f;
        float jaw_volume_scaling = 20f;

        /* MFCC parameters */
        const float fft_scaling = 4.4350e+06f;
        const float alpha = 0.97f;       
        const int M_FB = 20;            
        const int C = 12;               
        const int L = 22;              
        const int LF = 300;            // lower frequency limit(Hz)
        const int HF = 4500;           // upper frequency limit(Hz)
        int f_low;
        int f_high;
        float fres;

        float[,] filterBanks; 
        float[,] DCT_mat;
        float[] cep_lifter; 
        public float[] bs_setpoint;
        float[,] cc_mem;
        float bs_mouthOpen_setpoint;
        float current_volume;

        public override void Init()
        {
            pluginLabelJSON.val = "RT LipSync";
            viseme_morphs = new DAZMorph[16];
            JSONStorable js = containingAtom.GetStorableByID("geometry");
            if (js != null)
            {
                DAZCharacterSelector Dchar = js as DAZCharacterSelector;
                GenerateDAZMorphsControlUI morphUI = Dchar.morphsControlUI;
                if (morphUI != null)
                {
                    viseme_morphs[0] = morphUI.GetMorphByDisplayName("W");
                    viseme_morphs[1] = morphUI.GetMorphByDisplayName("UW");
                    viseme_morphs[2] = morphUI.GetMorphByDisplayName("TH");
                    viseme_morphs[3] = morphUI.GetMorphByDisplayName("T");
                    viseme_morphs[4] = morphUI.GetMorphByDisplayName("SH");
                    viseme_morphs[5] = morphUI.GetMorphByDisplayName("S");
                    viseme_morphs[6] = morphUI.GetMorphByDisplayName("OW");
                    viseme_morphs[7] = morphUI.GetMorphByDisplayName("M");
                    viseme_morphs[8] = morphUI.GetMorphByDisplayName("L");
                    viseme_morphs[9] = morphUI.GetMorphByDisplayName("K");
                    viseme_morphs[10] = morphUI.GetMorphByDisplayName("IY");
                    viseme_morphs[11] = morphUI.GetMorphByDisplayName("IH");
                    viseme_morphs[12] = morphUI.GetMorphByDisplayName("F");
                    viseme_morphs[13] = morphUI.GetMorphByDisplayName("ER");
                    viseme_morphs[14] = morphUI.GetMorphByDisplayName("EH");
                    viseme_morphs[15] = morphUI.GetMorphByDisplayName("AA");
                    mouthOpen_morph = morphUI.GetMorphByDisplayName("Mouth Open");
                    for (int i = 0; i < no_visemes; i++)
                    {
                        viseme_morphs[i].isTransient = false;
                        viseme_morphs[i].isRuntime = false;
                    }
                    mouthOpen_morph.isTransient = false;
                    mouthOpen_morph.isRuntime = false;
                }

                headAudioSource = Dchar.containingAtom.GetStorableByID("HeadAudioSource") as AudioSourceControl;
                voice_source = headAudioSource.audioSource;
            }

            JSONfloat_lipMovement = new JSONStorableFloat("Lip Movement Amount", 0.65f, 0.0f, 1.0f, false);
            RegisterFloat(JSONfloat_lipMovement);
            CreateSlider(JSONfloat_lipMovement, true); // true -> right side

            JSONfloat_jawMovement = new JSONStorableFloat("Jaw Movement Amount", 0.8f, 0.0f, 1.0f, false);
            RegisterFloat(JSONfloat_jawMovement);
            CreateSlider(JSONfloat_jawMovement, true); // true -> right side

            JSONfloat_bsUpdateSpeed = new JSONStorableFloat("Update Speed", 0.4f, 0.0f, 1.0f, false);
            RegisterFloat(JSONfloat_bsUpdateSpeed);
            CreateSlider(JSONfloat_bsUpdateSpeed, true); // true -> right side
            
        }

        void Start()
        {
            SuperController.LogMessage("RT LipSync started");
            //SuperController.LogMessage("directory: " + SuperController.singleton.currentLoadDir);

            bs_setpoint = new float[no_visemes];
            cc_mem = new float[C + 1, 2];
            bs_mouthOpen_setpoint = 0;

            filterBanks = LipSyncCalculation.CreateFilterBanks(fs, N_freq_bins, frequency_resolution, M_FB, LF, HF);
            cep_lifter = LipSyncCalculation.CreateCC_lifter(C + 1, L);
            DCT_mat = LipSyncCalculation.GenerateDCTmatrix(C + 1, M_FB);
            fres = fs / N_freq_bins / 2;
            f_low = UnityEngine.Mathf.RoundToInt(LF / fres);
            f_high = UnityEngine.Mathf.RoundToInt(HF / fres);
            
        }
        
        public void FixedUpdate()
        {
            volume_scaling = 500f * UnityEngine.Mathf.Exp(-6.111f * JSONfloat_lipMovement.val);
            jaw_volume_scaling = 500f * UnityEngine.Mathf.Exp(-6.111f * JSONfloat_jawMovement.val);
            bs_change_speed = JSONfloat_bsUpdateSpeed.val;

            if (voice_source.isPlaying)
            {
                float[] spectrum = new float[N_freq_bins];
                voice_source.GetSpectrumData(spectrum, 0, UnityEngine.FFTWindow.Hamming);

                f_low = UnityEngine.Mathf.RoundToInt(LF / fres);
                f_high = UnityEngine.Mathf.RoundToInt(HF / fres);

                float last_volume = current_volume;
                float[] audioTrace = new float[256];
                voice_source.GetOutputData(audioTrace, 0);
                current_volume = 0;
                for (int i = 0; i < 256; i++)
                    current_volume += Mathf.Abs(audioTrace[i]);
                current_volume = last_volume * (1 - bs_change_speed) + current_volume * bs_change_speed;

                float[] mfccs = LipSyncCalculation.MFCC(spectrum, filterBanks, DCT_mat, cep_lifter, ref cc_mem, M_FB, C, f_low, f_high);
                int VQregion = LipSyncCalculation.VQ_evaluate(mfccs, C, no_phonemes);

                for (int i = 0; i < no_visemes; i++)
                {
                    bs_setpoint[i] = VoiceProfile.influence[VQregion, i] * current_volume / volume_scaling;
                }
                bs_mouthOpen_setpoint = current_volume / jaw_volume_scaling * .1f;

                UpdateBlendshapes();
            }
        }

        void UpdateBlendshapes()
        {

            for (int i = 0; i < no_visemes; i++)
            {
                if (viseme_morphs[i] != null)
                {
                    float currentVisemeValue = viseme_morphs[i].morphValue;
                    if (currentVisemeValue > bs_setpoint[i])
                        viseme_morphs[i].SetValue(currentVisemeValue * (1 - bs_change_speed) + bs_setpoint[i] * bs_change_speed);
                    else
                        viseme_morphs[i].SetValue(currentVisemeValue * (1 - bs_change_speed/2) + bs_setpoint[i] * bs_change_speed/2);
                }
            }
            if (mouthOpen_morph != null)
            {
                float currentValue = mouthOpen_morph.morphValue;
                if (currentValue > bs_mouthOpen_setpoint)
                    mouthOpen_morph.SetValue(currentValue * (1 - bs_change_speed) + bs_mouthOpen_setpoint * bs_change_speed);
                else
                    mouthOpen_morph.SetValue(currentValue * (1 - bs_change_speed/2) + bs_mouthOpen_setpoint * bs_change_speed/2);

            }
        }

        protected JSONStorableFloat JSONfloat_lipMovement;
        protected JSONStorableFloat JSONfloat_bsUpdateSpeed;
        protected JSONStorableFloat JSONfloat_jawMovement;
    }

    public class LipSyncCalculation
    {

        public static float[] MFCC(float[] spectrum, float[,] filterBanks, float[,] DCT_mat, float[] cep_lifter, ref float[,] cc_mem, int M_FB, int C, int f_low, int f_high)
        {
            float[] FB_energy_log = new float[M_FB];
            float maxVal = -1f;

            for (int i = 0; i < spectrum.Length; i++)
                if (spectrum[i] > maxVal)
                    maxVal = spectrum[i];
            for (int i = 0; i < spectrum.Length; i++)
                spectrum[i] /= maxVal;

            for (int m = 0; m < M_FB; m++)
            {
                FB_energy_log[m] = 0;
                for (int i = f_low; i < f_high; i++)
                {
                    FB_energy_log[m] += filterBanks[m, i] * UnityEngine.Mathf.Abs(spectrum[i]);
                }
                FB_energy_log[m] = Mathf.Log(FB_energy_log[m]);
            }
            int N = C + 1;
            float[] c = new float[N];
            float[] c_lift = new float[N];

            for (int i = 0; i < N; i++)
            {
                c[i] = 0;
                for (int j = 0; j < M_FB; j++)
                {
                    c[i] += DCT_mat[i, j] * FB_energy_log[j];
                }
            }

            // Lift operation:
            for (int i = 0; i < N; i++)
            {
                c_lift[i] = c[i] * cep_lifter[i];
            }

            float[] c_full = new float[N * 2];

            for (int i = 0; i < N; i++)
            {
                c_full[i] = c_lift[i];
            }
            for (int i = N; i < 2 * N; i++)
            {
                c_full[i] = ((c_lift[i - N] - cc_mem[i - N, 1]) + 2 * (c_lift[i - N] - cc_mem[i - N, 0])) / 8f;
            }

            for (int i = 0; i < N; i++)
            {
                cc_mem[i, 0] = cc_mem[i, 1];
                cc_mem[i, 1] = c_lift[i];
            }

            return c_full;
        }
        
        public static float[] CreateCC_lifter(int N, int L)
        {
            float[] lifter = new float[N];

            for (int i = 0; i < N; i++)
            {
                lifter[i] = 1 + 0.5f * (float)L * Mathf.Sin(Mathf.PI * i / (float)L);
            }
            return lifter;
        }


        public static float[,] CreateFilterBanks(float fs, int N_freq_bins, float frequency_resolution, int M_FB, int LF, int HF)
        {
            float[] f = new float[N_freq_bins];
            for (int i = 0; i < N_freq_bins; i++)
                f[i] = i * frequency_resolution;

            float f_max = fs / 2;

            float[] f_fb_range = new float[N_freq_bins];
            for (int i = 0; i < N_freq_bins; i++)
                f_fb_range[i] = i * f_max / N_freq_bins;

            int[] filterCutoffs = new int[M_FB + 2];
            float filter_width = (Hertz2Mel(HF) - Hertz2Mel(LF)) / ((float)M_FB + 1f);
            for (int k = 0; k < (M_FB + 2); k++)
            {
                float cutoff_mel = Mel2Hertz(Hertz2Mel(LF) + k * filter_width);
                filterCutoffs[k] = Mathf.RoundToInt(cutoff_mel);
            }

            float[,] filterBanks = new float[M_FB, N_freq_bins];
            for (int m = 0; m < M_FB; m++)
                for (int i = 0; i < N_freq_bins; i++)
                    filterBanks[m, i] = 0;

            for (int m = 0; m < M_FB; m++)
            {
                // up slope
                for (int k = 0; k < N_freq_bins; k++)
                {
                    if ((f_fb_range[k] > filterCutoffs[m]) & (f_fb_range[k] < filterCutoffs[m + 1]))
                        filterBanks[m, k] = (f_fb_range[k] - (float)filterCutoffs[m]) / ((float)filterCutoffs[m + 1] - (float)filterCutoffs[m]);
                    else if ((f_fb_range[k] > filterCutoffs[m + 1]) & (f_fb_range[k] < filterCutoffs[m + 2]))
                        filterBanks[m, k] = ((float)filterCutoffs[m + 2] - f_fb_range[k]) / ((float)filterCutoffs[m + 2] - (float)filterCutoffs[m + 1]);
                }
            }
            return filterBanks;
        }

        public static float Mel2Hertz(float mel)
        {
            return 700 * Mathf.Exp(mel / 1127f) - 700f;
        }

        public static float Hertz2Mel(float hertz)
        {
            return 1127 * Mathf.Log(1 + hertz / 700f);
        }

        public static float[,] GenerateDCTmatrix(int N, int M)
        {
            float[,] DCTCoeff = new float[N, M];
            for (int i = 0; i < N; i++)
            {
                for (int j = 0; j < M; j++)
                {
                    float a = i;
                    float b = Mathf.PI * ((float)(j + 1f) - 0.5f) / M;
                    DCTCoeff[i, j] = Mathf.Sqrt(2.0f / (float)M) * Mathf.Cos(a * b);
                }
            }
            return (DCTCoeff);
        }

        public static int VQ_evaluate(float[] mfccs, int C, int no_phonemes)
        {

            int M = (C + 1) * 2;
            float minSize = 1e6f;
            int minRegion = -1;
            float[] VQ_distance = new float[no_phonemes];
            for (int region = 0; region < no_phonemes; region++)
            {
                float distance = 0;
                for (int m = 0; m < M; m++)
                {
                    distance += Mathf.Pow(mfccs[m] - VoiceProfile.VQ_center[m, region], 2.0f);
                }
                VQ_distance[region] = Mathf.Sqrt(distance);
                if (VQ_distance[region] < minSize)
                {
                    minSize = VQ_distance[region];
                    minRegion = region;
                }
            }
            return minRegion;
        }
    }




    public class VoiceProfile
    {

        public static float[,] VQ_center = {
    {1.7304f,1.4675f,-1.7908f,5.7335f,0.2831f,3.0228f,-16.8621f,1.1982f,-0.32185f,-0.55248f,2.7598f,3.2818f,5.6581f,1.969f,3.9626f,0.69628f,0.023416f,5.0097f,-1.3303f,7.0069f,3.6275f,0.22556f,-3.329f,3.2902f,3.5636f,3.1561f,6.6088f,4.6011f,3.4376f,2.593f,-3.2948f,-1.8511f,},
    {-7.0822f,-6.076f,-3.6586f,-11.5086f,-8.3847f,-6.5122f,-12.698f,-1.2411f,-8.9218f,-9.3606f,-8.2212f,-0.78911f,-6.2238f,-6.7511f,-6.447f,6.0164f,-5.9766f,-6.2576f,-6.4354f,-4.6609f,-9.7257f,-6.8444f,-14.9884f,1.1389f,-7.2984f,-6.4767f,-7.0984f,-4.6185f,-1.9809f,-6.6026f,-1.9533f,0.024086f,},
    {-2.3339f,-4.1328f,17.9487f,-8.198f,-3.4308f,1.0169f,2.3126f,-2.9523f,10.1806f,8.3452f,10.9087f,-7.1714f,0.81147f,0.78755f,4.8626f,3.4894f,-1.6601f,-1.3191f,17.6286f,-4.9555f,3.5426f,1.6158f,5.1887f,-2.1362f,-3.6219f,0.46515f,0.19393f,-8.6992f,0.21476f,0.73389f,4.4757f,1.1567f,},
    {-4.2372f,-4.8839f,4.7117f,-3.2734f,-6.1454f,-1.1373f,-1.7869f,-13.4514f,2.3848f,3.7117f,6.4252f,7.8416f,1.2636f,-0.51298f,8.7585f,-10.9353f,-2.602f,-2.1819f,-5.7583f,-4.5474f,0.88259f,-0.4736f,-1.9447f,-15.2548f,-0.32283f,-0.81429f,0.017565f,-7.9789f,3.0759f,0.50879f,4.4907f,5.3661f,},
    {0.82468f,2.0858f,-4.9971f,-0.63463f,-2.5284f,3.6665f,0.38019f,2.9329f,-7.114f,-1.6618f,-9.1919f,1.0555f,1.072f,3.8647f,-2.6273f,-1.0858f,2.8976f,-0.41933f,-1.9631f,-0.70228f,-9.1671f,4.0012f,-0.032453f,4.842f,-4.7524f,-1.2575f,-9.0476f,1.2495f,3.6624f,3.1795f,2.6712f,1.3138f,},
    {-1.1421f,-3.6843f,18.9604f,-3.8628f,-9.1097f,3.4715f,-1.5911f,-5.9128f,3.2379f,5.8058f,5.7832f,-14.1405f,-2.0219f,3.6075f,-3.6978f,6.3877f,1.1369f,-2.8172f,6.8162f,-5.371f,-5.7467f,3.7061f,-0.1144f,-0.39352f,-10.4552f,-2.8109f,-4.2785f,-9.2306f,-1.062f,4.0472f,-0.91735f,4.8309f,},
    {-0.75189f,3.6373f,-5.5182f,5.0561f,-11.9223f,8.9839f,1.6277f,-6.9286f,-5.6957f,3.6312f,1.1766f,5.1125f,-0.76923f,8.2207f,0.6426f,-0.16095f,3.9198f,1.9913f,-7.8147f,-5.1291f,-5.9605f,9.1443f,0.83358f,-3.0971f,-2.0375f,-0.11059f,3.6475f,-0.38399f,-8.4274f,8.636f,4.5487f,8.263f,},
    {-10.0219f,4.268f,0.032414f,2.2131f,-23.093f,9.6041f,-2.6489f,-7.6321f,-8.2064f,5.9749f,-5.1447f,-4.303f,0.35102f,6.3011f,-0.88565f,-5.4918f,-1.5981f,2.3608f,3.5488f,7.7182f,-9.2132f,10.8356f,-3.6505f,-2.5317f,-3.8089f,0.079044f,0.80476f,-2.2784f,-1.8623f,7.3353f,3.862f,-8.7673f,},
    {-19.5906f,8.0624f,1.1607f,-3.5108f,-19.752f,12.1016f,-1.1262f,-6.0761f,-7.9746f,7.8241f,1.4319f,-1.3991f,-2.922f,3.1415f,-7.8926f,-0.051729f,-10.1242f,2.7983f,-5.4093f,-1.6925f,-5.3365f,13.63f,-1.6623f,-2.4279f,-8.5496f,0.40102f,-2.8391f,7.9757f,2.1662f,5.7731f,-3.4749f,3.5819f,},
    {-24.6821f,11.9841f,-4.1373f,0.77607f,-5.2969f,12.3076f,0.3505f,-6.7054f,-10.6384f,9.8606f,-0.63361f,-2.1706f,-0.066022f,-5.0272f,-1.2361f,-5.3086f,-21.7635f,6.5561f,-1.161f,-1.4446f,-1.6603f,12.7763f,0.038066f,3.9055f,-7.6113f,2.6555f,3.924f,-0.97759f,-6.9931f,-1.92f,1.3338f,-2.6775f,},
    {-16.7096f,17.8536f,-2.6582f,1.958f,12.0956f,8.3827f,-0.58167f,-8.8667f,-9.7634f,12.7025f,-0.86558f,1.6302f,-0.041994f,-15.4476f,0.12419f,1.3857f,-25.8693f,8.2443f,-0.91217f,2.2309f,-0.41371f,6.727f,-1.6066f,-0.21294f,-6.6994f,5.5195f,-0.00012753f,-0.44625f,1.2371f,-11.3654f,0.24934f,-0.19955f,},
    {0.62051f,19.5787f,-2.5817f,0.072658f,10.3952f,-0.59011f,-0.90052f,-7.9665f,-9.6426f,13.0537f,0.29848f,0.30153f,-0.71287f,-19.4491f,-4.3129f,-1.9042f,-15.6755f,9.5154f,0.15993f,-0.52025f,1.0907f,-3.8141f,-0.67309f,1.1884f,-8.9414f,8.5137f,-0.68892f,0.69333f,-1.7957f,-15.9732f,-0.80992f,-3.1914f,},
    {6.1193f,14.2215f,0.63466f,-0.44046f,-1.9867f,-11.5248f,-0.42918f,-8.5563f,-8.1148f,9.43f,0.68374f,0.18404f,-0.75128f,-14.0619f,-0.12718f,0.471f,-0.62107f,8.3672f,0.48186f,0.43528f,3.0822f,-16.2822f,-0.51674f,0.60272f,-8.4116f,9.7711f,0.68036f,-0.51469f,0.8479f,-13.3107f,-0.83878f,-0.2367f,},
    {-0.91567f,-0.47202f,-0.70123f,0.25411f,-1.0894f,-0.31661f,-3.961f,-0.45881f,-0.80975f,-0.8624f,0.27103f,-0.21767f,10.9272f,-0.36422f,2.0454f,-0.84428f,0.039981f,1.2472f,-0.80628f,0.48092f,0.14921f,-0.082732f,-2.2992f,0.0014582f,-0.054063f,0.86484f,1.0169f,0.28468f,0.89555f,1.0622f,-3.3855f,-2.022f,},
    {0.50829f,-0.14069f,0.70884f,-0.6264f,0.23926f,0.427f,-1.6301f,0.80608f,0.20466f,0.021673f,-0.070559f,0.49454f,-1.5215f,0.56549f,1.1862f,1.9371f,0.31855f,0.72728f,0.43737f,-0.14281f,-0.545f,0.16957f,-3.3358f,1.2044f,-0.12433f,0.097974f,-0.68552f,-0.13745f,1.8633f,0.68015f,0.92013f,1.8452f,},
    {-0.52778f,-0.78256f,0.97476f,-1.2569f,0.18971f,0.40823f,-0.51011f,-0.83845f,1.6399f,1.2719f,3.0167f,-0.90739f,-4.7238f,0.15731f,0.89628f,1.2383f,0.10984f,-0.29607f,2.6535f,-0.60716f,0.72514f,0.46015f,0.032977f,-0.27996f,-0.78385f,-0.20835f,0.91732f,-1.6839f,-0.74596f,0.090327f,1.1069f,-1.0392f,},
    {-0.2417f,-0.46929f,0.99499f,-0.37847f,-0.27519f,1.1737f,-1.3099f,-2.3446f,0.48318f,0.67268f,1.335f,2.2015f,-1.3887f,0.60004f,3.4888f,-1.3833f,0.36211f,-0.034781f,-1.4331f,-0.079828f,0.43332f,-0.076317f,-1.5138f,-2.7157f,0.42315f,0.13612f,0.41952f,-0.8922f,0.22059f,0.6288f,1.2788f,0.62217f,},
    {-0.37702f,0.13688f,0.081188f,0.62897f,-0.68797f,1.4205f,0.22789f,0.92394f,-1.0805f,-0.23933f,-1.7651f,0.27869f,0.80733f,0.53645f,0.34552f,-1.4516f,0.81431f,-0.34517f,1.0486f,-0.24736f,-2.4765f,-0.39566f,0.29425f,1.4673f,-1.4103f,-0.70058f,-3.277f,0.44436f,2.5253f,1.1039f,2.4433f,2.0797f,},
    {-0.90793f,-0.49749f,2.5574f,0.40596f,-1.6762f,2.1352f,-0.01748f,-0.44416f,1.5718f,1.6796f,3.3949f,-2.8471f,-4.5273f,0.4113f,-1.5562f,2.0282f,1.3251f,-0.49095f,1.4784f,-0.83523f,-1.3964f,0.079167f,-0.17053f,0.68712f,-2.056f,-1.4408f,-0.13685f,-1.7131f,1.1453f,1.2622f,1.081f,1.7302f,},
    {-1.7405f,0.09321f,-0.41012f,0.86203f,-3.8732f,3.4921f,-0.082052f,-1.3372f,-0.93361f,0.2379f,0.37252f,1.7855f,-2.694f,0.75885f,0.68202f,2.0467f,2.7661f,0.76852f,-1.6844f,-0.8432f,-2.0575f,-0.37491f,0.18034f,-0.58335f,0.31595f,-2.0456f,1.7501f,-0.52571f,-3.7744f,2.1385f,2.2444f,3.4688f,},
    {-5.6029f,-0.091478f,0.72056f,0.41631f,-6.3096f,4.2597f,-0.52524f,-1.7126f,-1.3516f,1.3446f,-0.99652f,0.055054f,2.0384f,-1.001f,1.2677f,-0.33956f,3.7914f,2.4086f,2.2184f,2.6065f,-3.0415f,-0.061732f,-1.2837f,-1.3145f,-0.22447f,-3.497f,-0.48265f,0.7023f,-0.64658f,5.1616f,1.8139f,-2.8379f,},
    {-10.5465f,0.67858f,0.30677f,-0.6481f,-4.8029f,4.1873f,0.7294f,-0.52619f,-1.6541f,1.5163f,1.7837f,-0.019174f,-1.2876f,-5.1084f,-2.9362f,-0.54246f,3.2602f,3.417f,-1.0006f,-0.013678f,-0.75801f,0.54576f,-0.015568f,-0.20409f,-1.6782f,-5.2649f,0.53582f,3.8783f,2.4147f,8.6335f,0.92996f,1.2044f,},
    {-11.2173f,1.2043f,-0.80487f,-0.13659f,2.0219f,2.2778f,0.26672f,-1.1123f,-2.8527f,1.5959f,-0.22691f,-0.071188f,1.6695f,-9.6691f,-0.18192f,-1.1111f,-0.02114f,6.0302f,0.50348f,-0.50489f,0.26346f,2.0172f,0.52576f,1.7834f,-1.9088f,-6.2981f,2.2359f,0.66918f,-1.9593f,10.1767f,0.59791f,0.36354f,},
    {-3.8881f,2.9815f,0.24445f,0.46052f,8.2343f,-2.6003f,0.20733f,-1.991f,-2.7267f,2.2211f,-0.18835f,0.45098f,0.97212f,-13.2793f,0.71795f,0.67715f,-4.7264f,7.8952f,0.53722f,0.78225f,0.20106f,2.9884f,-0.42367f,0.12755f,-2.0468f,-7.147f,-0.33416f,-0.68596f,0.2518f,8.5816f,0.3419f,-0.54324f,},
    {6.6197f,3.7852f,-0.388f,0.11806f,4.7603f,-9.2948f,0.27612f,-2.0574f,-2.8475f,2.5522f,0.94065f,0.52414f,1.3928f,-10.1334f,-0.93593f,-0.076496f,-6.2544f,9.0122f,0.25332f,0.091337f,0.83653f,3.3019f,0.16196f,0.62473f,-3.0076f,-6.13f,0.31451f,0.52641f,0.15978f,2.2392f,0.36453f,-0.34881f,},
    {7.825f,2.8969f,0.27022f,-0.042973f,-2.3996f,-13.2687f,0.14273f,-2.2486f,-2.6834f,2.1934f,0.38087f,0.49564f,-0.054112f,-0.89387f,0.40993f,0.18909f,-1.9671f,7.6017f,0.88271f,0.14969f,1.4703f,1.8329f,0.16513f,0.48392f,-3.107f,-2.9675f,0.81428f,0.058913f,0.52267f,-4.6492f,-0.092878f,-0.11284f,},
    };

        public static float[,] influence =
        {
        {.00f, .00f, .00f, .00f, .00f, .00f, .30f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .50f, 1.0f}, // 1
        {.00f, .00f, .00f, .00f, .00f, .00f, .40f, .00f, .00f, .00f, .35f, .00f, .00f, .00f, .00f, .25f, 1.0f}, // 2 
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .30f, .20f, .00f, .00f, .40f, .20f, 1.0f}, // 3
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .10f, .80f, 1.0f}, // 4
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .30f, .50f, .00f, .00f, .00f, .30f, 1.0f}, // 5.
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .20f, .00f, .00f, .00f, .20f, .40f, 1.0f}, // 6
        {.00f, .00f, .00f, .00f, .00f, .90f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, 1.0f}, // 7
        {.00f, .00f, .00f, .00f, .00f, .00f, .10f, .00f, .80f, .00f, .00f, .00f, .00f, .00f, .00f, .10f, 1.0f}, // 8
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .25f, .00f, .00f, .40f, .00f, 1.0f}, // 9
        {.00f, .20f, .00f, .00f, .00f, .30f, .00f, .00f, .00f, .00f, .10f, .00f, .00f, .00f, .30f, .00f, 1.0f}, // 10
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .30f, .30f, .00f, 1.0f}, // 11
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .90f, .20f, .00f, 1.0f}, // 12
        {.00f, .00f, .80f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, 1.0f}, // 13
        {.15f, .00f, .00f, .00f, .20f, .60f, .30f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, 1.0f}, // 14
        {.00f, .00f, .30f, .50f, .00f, .00f, .00f, .00f, .00f, .20f, .00f, .00f, .00f, .00f, .00f, .00f, 1.0f}, // 15
        {.10f, .20f, .00f, .00f, .00f, .00f, .50f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, 1.0f}, // 16
        {.00f, .60f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .40f, .00f, .00f, .00f, 1.5f}, // 17
        {.20f, .00f, .00f, .00f, .30f, .00f, .00f, .70f, .00f, .00f, .00f, .00f, .00f, .00f, .10f, .10f, 1.0f}, // 18 
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .90f, .00f, .00f, .00f, .00f, .00f, 1.0f}, // 19
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .30f, .60f, 1.0f}, // 20
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .20f, .00f, .00f, .00f, .80f, .00f, 1.0f}, // 21
        {.00f, .00f, .00f, .00f, .00f, .00f, .20f, .00f, .00f, .00f, .30f, .00f, .00f, .00f, .00f, .40f, 1.0f}, // 22
        {.00f, .00f, .00f, .70f, .30f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, 1.0f}, // 23
        {.15f, .00f, .00f, .00f, .00f, .00f, .30f, .20f, .00f, .20f, .00f, .00f, .00f, .00f, .00f, .30f, 1.0f}, // 24
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .90f, .00f, 1.0f}, // 25
        {.00f, .00f, .00f, .00f, .00f, .00f, .20f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .20f, .40f, 1.0f}, // 26
        {.00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .70f, .00f, .00f, .00f, .30f, .00f, 1.0f}, // 27
        {.20f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .20f, .00f, .30f, .40f, .00f, .30f, 1.0f}, // 28
        {.00f, .00f, .60f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .30f, .00f, .00f, .20f, .00f, .00f, 1.0f}, // 29
        {.00f, .00f, .00f, .10f, .00f, .20f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .30f, .00f, .30f, 1.0f}, // 30
        {.20f, .00f, .00f, .00f, .00f, .20f, .00f, .50f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, 1.5f}, // 31
        {.20f, .00f, .00f, .00f, .00f, .00f, .00f, .90f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, .00f, 1.1f}, // 32

    };

    }


}
